//////////////////////////////////////////////////////////////////////////
// UVRip 1.05 by Bendy's 3D Systems
// Copyright (c) 1999
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// The complete GNU public license can be found in this directory.
//
// Explicit permission is given to link this file with the 
// tsxapi.lib library supplied with the trueSpace API
// Software Developers Kit
//////////////////////////////////////////////////////////////////////////
/*  FILENAME: AALines.c  [revised 17 AUG 90]

    AUTHOR:  Kelvin Thompson

    DESCRIPTION:  Code to render an anti-aliased line, from
      "Rendering Anti-Aliased Lines" in _Graphics_Gems_.

      This is almost exactly the code printed on pages 690-693
      of _Graphics_Gems_.  An overview of the code is on pages
      105-106.

    LINK WITH:
      AALines.h -- Shared tables, symbols, etc.
      AAMain.c -- Calling code for this subroutine.
      AATables.c -- Initialize lookup tables.
*/

// Note: I think there may be some bugs in this code
// when used with complicated UV maps

// Implementation of antialiased line drawing. Slightly modified
// to work a bit more nicely with UVRip.
// Changes marked by // CHANGED

// CHANGED 4/4/99
// Now works on rgba colours

#include "stdafx.h"
#include "AALines.h"
#include <math.h>

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

#define SWAPVARS(v1,v2) ( temp=v1, v1=v2, v2=temp )

#define FIXMUL(f1,f2) \
  ( \
  (((f1&0x0000ffff) * (f2&0x0000ffff)) >> 16) + \
  ((f1&0xffff0000)>>16) * (f2&0x0000ffff) + \
  ((f2&0xffff0000)>>16) * (f1&0x0000ffff) + \
  ((f1&0xffff0000)>>16) * (f2&0xffff0000) \
  )

/* HARDWARE ASSUMPTIONS:
/*    * 32-bit, signed ints
/*    * 8-bit pixels, with initialized color table
/*    * pixels are memory mapped in a rectangular fashion */

/* FIXED-POINT DATA TYPE */
#ifndef FX_FRACBITS
  typedef int FX;
# define FX_FRACBITS 16  /* bits of fraction in FX format */
# define FX_0        0   /* zero in fixed-point format */
#endif

/* ASSUMED MACROS:
/*   SWAPVARS(v1,v2) -- swaps the contents of two variables
/*   PIXADDR(x,y) -- returns address of pixel at (x,y)
/*   COVERAGE(FXdist) -- lookup macro for pixel coverage 
        given perpendicular distance; takes a fixed-point
        integer and returns an integer in the range [0,255]
/*   SQRTFUNC(FXval) -- lookup macro for sqrt(1/(1+FXval^2))
        accepts and returns fixed-point numbers
/*   FIXMUL(FX1,FX2) -- multiplies two fixed-point numbers
        and returns the product as a fixed-point number   */

/* BLENDING FUNCTION:
/*  'cover' is coverage -- in the range [0,255]
/*  'back' is background color -- in the range [0,255] */
//#define BLEND(cover,back) ((((255-(cover))*(back))>>8)+(cover))

// CHANGED 5/5/98
// The above BLEND function works in greyscale only
// and the color is set to white.
// Below is my function which blends color into back
// by amount cover.
rgba
BLEND(int cover,rgba color,rgba back)
{
	color.blend(back,cover/255.0f);
	return color;
}
/* LINE DIRECTION bits and tables */
#define DIR_STEEP  1  /* set when abs(dy) > abs(dx) */
#define DIR_NEGY   2  /* set whey dy < 0 */

/* Global 'Pmax' is initialized elsewhere.  It is the
   "maximum perpendicular distance" -- the sum of half the
   line width and the effective pixel radius -- in fixed format */
FX Pmax;


/***************  FUNCTION ANTI_LINE  ***************/

// CHANGED 5/5/98
// Anti_Line now accepts the frame buffer, line color
// and frame resolution as parameters.
// CHANGED 23/4/99
// Added check to make sure that pixels being blended are
// within the assigned image.
// PRE CONDITIONS:
//					fbuff already has xpix*ypix RGB
//					pixels assigned to it.
void Anti_Line (rgba *fbuff, rgba color, int xpix, int ypix,
				int X1, int Y1, int X2, int Y2 )
{
int 	Bvar, 	/* decision variable for Bresenham's */
    	Bainc,   /* adjacent-increment for 'Bvar' */
    	Bdinc;   /* diagonal-increment for 'Bvar' */
FX 		Pmid,  	/* perp distance at Bresenham's pixel */
   		Pnow,  	/* perp distance at current pixel (ortho loop) */
   		Painc, 	/* adjacent-increment for 'Pmid' */
   		Pdinc, 	/* diagonal-increment for 'Pmid' */
   		Poinc; 	/* orthogonal-increment for 'Pnow'--also equals 'k' */
rgba 	*mid_addr,   /* pixel address for Bresenham's pixel */
     	*now_addr,   /* pixel address for current pixel */
		*end_image=fbuff+xpix*ypix;
int 	addr_ainc,   /* adjacent pixel address offset */
    	addr_dinc,   /* diagonal pixel address offset */
    	addr_oinc;   /* orthogonal pixel address offset */
int dx,dy,dir;    	/* direction and deltas */
FX slope;			/* slope of line */
int temp;

if (!coverage || !sqrtfunc)
	return;

/* rearrange ordering to force left-to-right */
if 	( X1 > X2 )
  	{ SWAPVARS(X1,X2);  SWAPVARS(Y1,Y2); }

/* init deltas */
dx = X2 - X1;  /* guaranteed non-negative */
dy = Y2 - Y1;

// CHANGED 5/5/98
// As TrueSpace has some nasty UV values we must check
// that this line isn't a point.
// If it is a point, draw it and return
if (dx==0 && dy==0)
{
	mid_addr = PIXADDR(X1,Y1);
	if (mid_addr >= fbuff && mid_addr < fbuff+xpix*ypix)
		*mid_addr = BLEND( COVERAGE(abs(FX_0)), color, *mid_addr );
	return;
}

/* pixel increment values 
/*   -- assume PIXINC(dx,dy) is a macro such that:
/*   PIXADDR(x0,y0) + PIXINC(dx,dy) = PIXADDR(x0+dx,y0+dy)  */
int adj_pixinc[4] = 
      { PIXINC(1,0), PIXINC(0,1), PIXINC(1,0), PIXINC(0,-1) };
int diag_pixinc[4] = 
      { PIXINC(1,1), PIXINC(1,1), PIXINC(1,-1), PIXINC(1,-1) };
int orth_pixinc[4] = 
      { PIXINC(0,1), PIXINC(1,0), PIXINC(0,-1), PIXINC(1,0) };

/* calculate direction (slope category) */
dir = 0;
if ( dy < 0 )   { dir |= DIR_NEGY;  dy = -dy; }
if ( dy > dx )  { dir |= DIR_STEEP; SWAPVARS(dx,dy); }


/* init address stuff */
mid_addr = PIXADDR(X1,Y1);
addr_ainc = adj_pixinc[dir];
addr_dinc = diag_pixinc[dir];
addr_oinc = orth_pixinc[dir];

/* perpendicular measures */
slope =  (dy << FX_FRACBITS) / dx;
Poinc = SQRTFUNC( slope );
Painc = FIXMUL( slope, Poinc );
Pdinc = Painc - Poinc;
Pmid = FX_0;

/* init Bresenham's */
Bainc = dy << 1;
Bdinc = (dy-dx) << 1;
Bvar = Bainc - dx;

do
  	{
  	/* do middle pixel */
  	if (mid_addr >= fbuff && mid_addr < fbuff+xpix*ypix)
		*mid_addr = BLEND( COVERAGE(abs(Pmid)), color, *mid_addr );

  	/* go up orthogonally */
  	for (
      	Pnow = Poinc-Pmid,  now_addr = mid_addr+addr_oinc;
      	Pnow < Pmax;
      	Pnow += Poinc,      now_addr += addr_oinc
      	)
       if (now_addr<end_image && now_addr>fbuff && (Pnow>>covershift)>0)
		   if (now_addr >= fbuff && now_addr < fbuff+xpix*ypix)
				*now_addr = BLEND( COVERAGE(Pnow), color, *now_addr );

  	/* go down orthogonally */
  	for (
      	Pnow = Poinc+Pmid,  now_addr = mid_addr-addr_oinc;
      	Pnow < Pmax;
      	Pnow += Poinc,      now_addr -= addr_oinc
      	)
		if (now_addr<end_image && now_addr>fbuff && (Pnow>>covershift)>0)
			if (now_addr >= fbuff && now_addr < fbuff+xpix*ypix)
				*now_addr = BLEND( COVERAGE(Pnow), color, *now_addr );


  	/* update Bresenham's */
  	if ( Bvar < 0 )
    	{
    	Bvar += Bainc;
    	mid_addr = (rgba *) (mid_addr + addr_ainc);
    	Pmid += Painc;
    	}
  	else
    	{
    	Bvar += Bdinc;
    	mid_addr = (rgba *) (/*(int)*/mid_addr + addr_dinc);
    	Pmid += Pdinc;
    	}

  	--dx;
  	} while ( dx >= 0 );
}
